/* 	Copyright John D. "smeezekitty" 2017
	
	FreeNBTV is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeNBTV is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeNBTV.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This file handles the channel setup GUI */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <math.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "nbtv_format.h"
#include "gui.h"
#include "input.h"
#include "nbtv_gen.h"
#include "util.h"

char file_picker_used = 0;
char updated_fn = 0;

uint8_t i_handler_disable; //Disable value change callbacks

GtkWidget *window_frame;

GtkWidget *input_window=NULL;
GtkWidget *grid;
GtkWidget *file_frame;
GtkWidget *windowing_frame;
GtkWidget *source_grid;
GtkWidget *windowing_grid;
GtkWidget *file_path;
GtkWidget *choose;
GtkWidget *windowing;
GtkWidget *rotation;
GtkWidget *label1,*label2,*label3;
GtkWidget *Xeb,*Yeb;
GtkWidget *iaspect;
GtkWidget *loop;

GtkWidget *p1,*p2,*p3;
GtkWidget *start;

GtkWidget *lock_aspect_os, *lock_aspect_op, *lock_aspect_u, *disc_corr;

/*File picker*/
void chooser_dialog(GtkWidget *button, gpointer window){
	GtkWidget *chooser;
	chooser = gtk_file_chooser_dialog_new("Select image/video", GTK_WINDOW(window), GTK_FILE_CHOOSER_ACTION_OPEN, "_OK", GTK_RESPONSE_OK, "_CANCEL", GTK_RESPONSE_CANCEL, NULL);
	gtk_widget_show_all(chooser);
	gint resp = gtk_dialog_run(GTK_DIALOG(chooser));
	if(resp == GTK_RESPONSE_OK){
		strcpy(input.filename, gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(chooser)));
		g_print("%s\n", input.filename);
		file_picker_used = 1;
		updated_fn = 1;
		set_file_path_box(input.filename);

	}
	gtk_widget_destroy(chooser);
}
void file_path_change(GtkEditable *editable, gpointer bla){
	updated_fn = 1;
	if(file_picker_used){
		set_file_path_box(input.filename);
		file_picker_used = 0;
		return;
	}
	strcpy(input.filename, gtk_editable_get_chars(editable, 0, 255));
	debug("%s\n", input.filename);
}
void set_file_path_box(char *fp){
	gtk_entry_set_text(GTK_ENTRY(file_path), fp);
}

void create_window_geometry_box(GtkWidget *grid, int xoff, int yoff, int width, int height){
	if(window_frame)gtk_widget_destroy(window_frame);
	window_frame = gtk_frame_new("Input windowing: ");
	gtk_grid_attach(GTK_GRID(grid), window_frame, xoff, yoff, width, height);
	
}
void set_window_rect(){
	input.window_type = 0;
}
void set_window_trap(){
	input.window_type = 1;
	if(!input.h1 || input.h1 > input.h2){
		input.h1 = input.h2 * 0.80;
	}
}
void set_window_annular(){
	input.window_type = 2;
}
void change_window_type(GtkWidget *cb, gpointer d){
	if(i_handler_disable) return;
	char *s = gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(cb));
	debug("%s\n", s);
	switch(*s){
		case 'R':
			set_window_rect();
			break;
		case 'T':
			set_window_trap();
			break;
		case 'A':
			set_window_annular();
			break;
	}
	format_changed = 1;
	gui_update_input();
}
void change_rotation_angle(GtkWidget *cb, gpointer d){
	if(i_handler_disable) return;
	input.rotation_angle = gtk_combo_box_get_active(GTK_COMBO_BOX(cb));
	double ratio=(double)input.h1/((double)input.h2+0.1);
	if(input.rotation_angle & 1){ //+90 or +270 degrees (so width and height are swapped)
		if(input.h2 > input.Xres){ //Is the "height" greater than the image width
			input.h2 = input.Xres; //Make it equal
		}
//		if(input.h1 > input.h2){
			input.h1 = input.h2 * ratio; //If it is trapezoidal, then fix the other base
//		}
		gui_update_input();
		update_p1(p1, NULL); //Just call the callback outselves to fix aspect ratio stuff. This doesn't work correctly and needs to be fixed
		if(input.w > input.Yres){	//Is the "width" greater than the height of the image
			input.w = input.Yres;
		}
	} else { //+0 or +180 degrees
		if(input.h2 > input.Yres){ //Is the width > image width
			input.h2 = input.Yres;	//Fix it
		}
//		if(input.h1 > input.h2){
			input.h1 = input.h2 * ratio;	//Ditto above
//		}
		gui_update_input();
		update_p1(p1, NULL);
		if(input.w > input.Xres){ //Is the height > image height
			input.w = input.Xres; //Fix it
		}
	}
	format_changed = 1;
	gui_update_input();
}
void update_input_xy(GtkWidget *v, gpointer w){
	if(i_handler_disable) return;
	if(w){
		input.y = gtk_spin_button_get_value(GTK_SPIN_BUTTON(v));
	} else {
		input.x = gtk_spin_button_get_value(GTK_SPIN_BUTTON(v));
	}
	format_changed = 1;
	//gui_update_input();
}
void aspect_lock_toggle(GtkWidget *toggle, gpointer w){
	if(i_handler_disable) return;
	/*This emulates radio buttons in the sense that only one checkbox can be checked. But it also leaves the option of have zero checked*/
	if(w == (gpointer)1){
		input.lock_aspect_to_pixel_format = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle));
		if(input.lock_aspect_to_pixel_format){input.lock_aspect_to_specified = input.lock_aspect_to_screen_format = 0;}
	} else if(w == 0){
		input.lock_aspect_to_screen_format = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle));
		if(input.lock_aspect_to_screen_format){input.lock_aspect_to_specified = input.lock_aspect_to_pixel_format = 0;}
	} else {
		input.lock_aspect_to_specified = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle));
		if(input.lock_aspect_to_specified){input.lock_aspect_to_screen_format = input.lock_aspect_to_pixel_format = 0;}
	}
	if(input.lock_aspect_to_pixel_format || input.lock_aspect_to_screen_format || input.lock_aspect_to_specified && input.window_type != 2){
		/*Enforce the aspect ratio if one of the aspect lock options are on*/
		if((double)input.w / (double)input.Xres > (double)input.h2 / (double)input.Yres){ /*If the width is closer to the width of the image*/
			double factor = (double)input.h1/(double)input.h2;
			if(floor(input.w/input.input_aspect) <= input.Yres){
				input.h2 = input.w/input.input_aspect;	/*Then change the height*/
			}
			if(input.window_type == 1){
					input.h1 = (double)input.h2 * factor + 0.5;
			}
		} else {
			if(floor(input.h2*input.input_aspect) <= input.Xres){
				input.w = input.h2*input.input_aspect;	/*Otherwise change the width*/
			}
		}
		/*If the current X,Y coordinants put it over the edge of the image, change the coordinants*/
		if(input.x + input.w/2 > input.Xres){
			input.x = input.Xres - input.w/2;	
		}
		if(input.y + input.h2/2 > input.Yres){
			input.y = input.Yres - input.h2/2;
		}
		format_changed = 1;
	}
	gui_update_input();
}
void loop_toggle(GtkWidget *lt, gpointer bla){
	input.loop = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lt));
	format_changed = 1;
}
/*Callback for parameter 1 (Width, Span or Theta)*/
void update_p1(GtkWidget *p, gpointer bla){
	int nw;
	if(i_handler_disable) return;
	if(input.window_type == 2) return;
	nw = gtk_spin_button_get_value(GTK_SPIN_BUTTON(p));
	double factor = (double)input.h1/(double)input.h2;
	if(input.lock_aspect_to_pixel_format || input.lock_aspect_to_screen_format || input.lock_aspect_to_specified){ 
		if(input.rotation_angle & 1){
			if(nw/input.input_aspect > input.Xres){
				gui_update_input();
				return;
			}
		} else {
			if(nw/input.input_aspect > input.Yres){
				gui_update_input();
				return;
			}
		}
		input.w = nw;
		input.h2 = input.w/input.input_aspect;
		if(input.window_type == 1){
			input.h1 = (double)input.h2 * factor + 0.5;	/*Change the height to keep the aspect ratio if we're doing that*/
		}
	} else {
		input.w = nw;
	}
	if(input.rotation_angle & 1){
		if(input.x + input.h2/2 > input.Xres){
			input.x = input.Xres - input.h2/2;
		}
		//if(input.x < input.h2/2){
//			input.x = input.h2 / 2;
	//	}
	} else {
		if(input.x + input.w/2 > input.Xres){
			input.x = input.Xres - input.w/2;
		}
		//if(input.x < input.w/2){
//			input.x = input.w/2;
	//	}
	}
	gui_update_input();
	format_changed = 1;
}
/*Callback for parameter 2 (Height, Base 2, Outer Radius)*/
void update_p2(GtkWidget *p, gpointer bla){
	if(i_handler_disable) return;
	if(input.window_type == 2) return;
	input.h2 = gtk_spin_button_get_value(GTK_SPIN_BUTTON(p));
	if(!input.h1){input.h1 = input.h2 * 0.80;}
	if(input.lock_aspect_to_pixel_format || input.lock_aspect_to_screen_format || input.lock_aspect_to_specified){
		input.w = input.h2*input.input_aspect;
	}
	if(input.rotation_angle & 1){
		if(input.y + input.w/2 > input.Yres){
			input.y = input.Yres - input.w/2;
		}
		//if(input.y < input.w/2){
			//input.y = input.w/2;
		//}
	} else {
		if(input.y + input.h2/2 > input.Yres){
			input.y = input.Yres - input.h2/2;
		}
		//if(input.y < input.h2/2){
//			input.y = input.h2/2;
	//	}
	}
	gui_update_input();
	format_changed = 1;
}
/*Callback for parameter 3 (Base 1, Inner Radius)*/
void update_p3(GtkWidget *p, gpointer bla){
	if(i_handler_disable) return;
	if(input.window_type == 1){
		input.h1 = gtk_spin_button_get_value(GTK_SPIN_BUTTON(p));
	}
	gui_update_input();
	format_changed = 1;
}

GtkWidget *create_rotation_combobox(){
	GtkWidget *cb = gtk_combo_box_text_new();
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "0", "+0\302\260");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "90", "+90\302\260");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "180", "+180\302\260");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "270", "+270\302\260");
	gtk_combo_box_set_active(GTK_COMBO_BOX(cb), 0);
	return cb;
}
GtkWidget *create_windowing_combobox(){
	GtkWidget *cb = gtk_combo_box_text_new();
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "r", "Rectangular");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "t", "Trapezoidal");
//	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), "a", "Annular"); //-- not yet implemented
	gtk_combo_box_set_active(GTK_COMBO_BOX(cb), 0);
	set_window_rect();
	return cb;
}
void gui_update_input(){
	char buffer[8];
	int x,y;
	if(!input_window) return;
	i_handler_disable=1;
	set_file_path_box(input.filename);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(loop), input.loop);
	gtk_widget_set_sensitive(iaspect, !(input.lock_aspect_to_screen_format|input.lock_aspect_to_pixel_format));
	gtk_widget_set_sensitive(p3, input.window_type);
	/*
	Set up the minimum and maximum range for the p1 and p2 spin buttons
	If the aspect locks are off, this is just the deminsions of the image.
	But if aspect ratio locks are enabled, make sure changing one won't make the other go over the size of the image
	*/
	if(input.rotation_angle & 1){ //+90,+270
			x = input.Yres;
			y = input.Xres;
			if(input.lock_aspect_to_screen_format || input.lock_aspect_to_pixel_format || input.lock_aspect_to_specified){
				x = min(x,input.Xres/input.input_aspect+1);
				y = min(y,input.Yres*input.input_aspect+1);
			}
		} else { //+0,+180
			x = input.Xres;
			y = input.Yres;
			if(input.lock_aspect_to_screen_format || input.lock_aspect_to_pixel_format || input.lock_aspect_to_specified){
				x = min(x,input.Yres*input.input_aspect+1);
				y = min(y,input.Xres/input.input_aspect+1);
			}
	}
	gtk_spin_button_set_range(GTK_SPIN_BUTTON(p1), 1, x);
	gtk_spin_button_set_range(GTK_SPIN_BUTTON(p2), 1, y);
	/////////////////
	/*Set up the spin button ranges for the X and Y coordinants so that it won't go past the end of the image for a given size*/
	if(input.rotation_angle & 1){
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(Xeb), input.h2/2, input.Xres - input.h2/2);
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(Yeb), input.w/2, input.Yres - input.w/2);
		frange(&input.x, input.h2/2, input.Xres - input.h2/2);
		frange(&input.y, input.w/2, input.Yres - input.w/2);
	} else {
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(Xeb), input.w/2, input.Xres - input.w/2);
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(Yeb), input.h2/2, input.Yres - input.h2/2);
		frange(&input.y, input.h2/2, input.Yres - input.h2/2);
		frange(&input.x, input.w/2, input.Xres - input.w/2);
	}
	if(input.window_type == 1){
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(p3), 1, input.h2);
		debug("Range set to 1 to %d\n", input.h2);
	} else if(input.window_type == 2){
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(p3), 0, 359);
	}
	////////////////
	if(input.window_type < 2){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(p1), input.w);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(p2), input.h2);
	}
	if(input.window_type == 0){ /*Rectangular. P1 = width, P2 = Height, P3 disabled*/
		gtk_label_set_text(GTK_LABEL(label1), "Width:");
		gtk_label_set_text(GTK_LABEL(label2), "Height:");
		gtk_label_set_text(GTK_LABEL(label3), "");
	} else if(input.window_type == 1){ /*Trapezoidal. P1 = span (~width), P2 = Base 2 (~height), P3 = Base 1 (The smaller leg)*/
		gtk_label_set_text(GTK_LABEL(label1), "Span:");
		gtk_label_set_text(GTK_LABEL(label2), "Base 2:");
		gtk_label_set_text(GTK_LABEL(label3), "Base 1:");
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(p3), input.h1);
	//	printf("Setting range of p3 to %d to %d\n", 1, input.h2);
	//	gtk_entry_set_text(GTK_ENTRY(p3), itoa(input.h1, buffer, 10));
	} else if(input.window_type == 2){ /*Annular. P1 = theta, P2 = Outer Radius, P3 = Inner Radius*/
		gtk_label_set_text(GTK_LABEL(label1), "\xce\xb8"); //Theta
		gtk_label_set_text(GTK_LABEL(label2), "Outer\nRadius:");
		gtk_label_set_text(GTK_LABEL(label3), "Innter\nRadius:");
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(p1), input.r2);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(p2), input.r1);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(p3), input.angle);
	}
	if(input.lock_aspect_to_screen_format){
		input.input_aspect = nbtv_fmt->aspect_ratio;
	} else if(input.lock_aspect_to_pixel_format){
		input.input_aspect = (double)nbtv_fmt->lines/(double)nbtv_fmt->pixels_per_line;
	} else if(input.lock_aspect_to_specified){
		input.input_aspect = atof(gtk_entry_get_text(GTK_ENTRY(iaspect)));
		if(input.input_aspect < 0.05 || input.input_aspect > 32) input.input_aspect = 1;
	}
	gtk_combo_box_set_active(GTK_COMBO_BOX(windowing), input.window_type);
	gtk_combo_box_set_active(GTK_COMBO_BOX(rotation), input.rotation_angle);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lock_aspect_os), input.lock_aspect_to_screen_format);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lock_aspect_op), input.lock_aspect_to_pixel_format);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lock_aspect_u), input.lock_aspect_to_specified);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(Xeb), input.x);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(Yeb), input.y);
	gtk_entry_set_text(GTK_ENTRY(iaspect), gcvt(input.input_aspect, 4, buffer));
	i_handler_disable = 0;
}

void gui_input_connect_callbacks(){
	g_signal_connect(file_path, "activate", G_CALLBACK(file_path_change), NULL);
	g_signal_connect(choose, "clicked", G_CALLBACK(chooser_dialog), NULL);
	g_signal_connect(windowing, "changed", G_CALLBACK(change_window_type), NULL);
	g_signal_connect(rotation, "changed", G_CALLBACK(change_rotation_angle), NULL);
	g_signal_connect(Xeb, "value-changed", G_CALLBACK(update_input_xy), NULL);
	g_signal_connect(Yeb, "value-changed", G_CALLBACK(update_input_xy), (gpointer)1);
	g_signal_connect(p1, "value-changed", G_CALLBACK(update_p1), NULL);
	g_signal_connect(p2, "value-changed", G_CALLBACK(update_p2), NULL);
	g_signal_connect(p3, "value-changed", G_CALLBACK(update_p3), NULL);
	g_signal_connect(lock_aspect_os, "toggled", G_CALLBACK(aspect_lock_toggle), NULL);
	g_signal_connect(lock_aspect_op, "toggled", G_CALLBACK(aspect_lock_toggle), (gpointer)1);
	g_signal_connect(lock_aspect_u, "toggled", G_CALLBACK(aspect_lock_toggle), (gpointer)2);
	g_signal_connect(loop, "toggled", G_CALLBACK(loop_toggle), NULL);
	
}

GtkWidget *create_input_window(){
	input_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect (input_window, "destroy",G_CALLBACK (destroy), NULL);
	gtk_window_set_title(GTK_WINDOW(input_window), "Input");
	gtk_window_set_resizable(GTK_WINDOW(input_window), FALSE);
	grid = gtk_grid_new();
	gtk_container_add(GTK_CONTAINER(input_window), grid);
	file_frame = gtk_frame_new("Source: ");
	gtk_grid_attach(GTK_GRID(grid), file_frame, 0, 0, 1, 1);
	source_grid = gtk_grid_new();
	gtk_container_add(GTK_CONTAINER(file_frame), source_grid);
	file_path = gtk_entry_new();
	gtk_grid_attach(GTK_GRID(source_grid), file_path, 1, 0, 1, 1);
	put_label(source_grid, "Image/\nVideo: ", 0, 0);
	choose = gtk_button_new_with_label("...");
	gtk_grid_attach(GTK_GRID(source_grid), choose, 2, 0, 1, 1);
	windowing_frame = gtk_frame_new("Windowing: ");
	windowing_grid = gtk_grid_new();
	gtk_grid_set_row_spacing(GTK_GRID(windowing_grid), 5);
	gtk_container_add(GTK_CONTAINER(windowing_frame), windowing_grid);
	gtk_grid_attach(GTK_GRID(grid), windowing_frame, 0, 1, 1, 1);
	Xeb = gtk_spin_button_new_with_range(0, 32, 1);
	Yeb = gtk_spin_button_new_with_range(0, 48, 1);
	put_label(windowing_grid, "X: ", 0, 1);
	put_label(windowing_grid, "Y: ", 0, 2);
	gtk_grid_attach(GTK_GRID(windowing_grid), Xeb, 1, 1, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), Yeb, 1, 2, 1, 1);
	windowing = create_windowing_combobox();
	gtk_grid_attach(GTK_GRID(windowing_grid), windowing, 0, 0, 2, 1);
	rotation = create_rotation_combobox();
	gtk_grid_attach(GTK_GRID(windowing_grid), rotation, 2, 0, 1, 1);
	label1 = gtk_label_new("Width:");
	label2 = gtk_label_new("Height:");
	label3 = gtk_label_new("C");
	gtk_grid_attach(GTK_GRID(windowing_grid), label1, 0, 4, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), label2, 0, 5, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), label3, 0, 6, 1, 1);
	p1 = gtk_spin_button_new_with_range(0, 32, 1);
	p2 = gtk_spin_button_new_with_range(0, 48, 1);
	p3 = gtk_spin_button_new_with_range(0, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), p1, 1, 4, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), p2, 1, 5, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), p3, 1, 6, 1, 1);
	iaspect = create_number_box(4);
	gtk_grid_attach(GTK_GRID(windowing_grid), iaspect, 1, 7, 1, 1);
	//GtkWidget *lock_aspect_os, *lock_aspect_op, *lock_aspect_u;
	loop = gtk_check_button_new_with_label("Loop");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(loop), 1);
	input.loop = 1;
	lock_aspect_os = gtk_check_button_new_with_label("Lock Aspect\nTo Screen Fmt");
	lock_aspect_op = gtk_check_button_new_with_label("Lock Aspect\nTo Pixel Fmt");
	lock_aspect_u = gtk_check_button_new_with_label("Lock Aspect\nTo Specified");
	disc_corr = gtk_check_button_new_with_label("Use Disc\nCorrection");
	gtk_widget_set_sensitive(disc_corr, 0);
	gtk_grid_attach(GTK_GRID(source_grid), loop, 0, 1, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), lock_aspect_os, 2, 4, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), lock_aspect_op, 2, 5, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), lock_aspect_u, 2, 6, 1, 1);
	gtk_grid_attach(GTK_GRID(windowing_grid), disc_corr, 2, 7, 1, 1);
	put_label(windowing_grid, "Aspect\nRatio:", 0, 7);
	gui_input_connect_callbacks();
	gtk_widget_show_all(input_window);
	return input_window;
}

