/* 	Copyright John D. "smeezekitty" 2017
	
	FreeNBTV is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeNBTV is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeNBTV.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This file is for the preview image and overlays*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "nbtv_format.h"
#include "gui.h"
#include "input.h"
#include "preview.h"
#include "nbtv_gen.h"
#include "util.h"

//SDL_Surface *input_preview;

//Performs an xor (color inversion) if it results in a visible change, otherwise it increases the gain
void overlay_alter_pixels(uint8_t *pixels, int start){
	uint8_t n;
	uint16_t o,p;
	for(n = 0;n < 3;n++){
		o = pixels[start + n];
		p = o ^ 0XFF;
		if(abs(o - p) < 32){
			p = o + 64;
			if(p > 255){
				p = 255;
			}
		}
		pixels[start + n] = p;
	}
}
//Draw rectangle rotated (90 or 270 degrees)
void draw_rectangular_overlay_r(SDL_Surface *s, struct _input *i){
	int x,y,p,n,m;
	int pixel_index;
	int pixel_offset;
	int max_addr = (s->h + 1) * s->pitch;
	uint8_t *pixels = (uint8_t*)s->pixels;
	int ybase = i->y*i->YRR - i->w*i->XRR/2;
	int xbase = i->x*i->XRR - i->h2*i->YRR/2;
	y = ybase;
	m = min(5, i->h2*i->YRR);
	for(x = xbase;x < (i->h2*i->YRR)+xbase && y < s->h;x++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = ((int)((i->w * i->XRR) + p)) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	x = xbase;
	m = min(5, i->w*i->XRR);
	for(y = ybase;y < (i->w*i->XRR)+ybase;y++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = (p + (i->h2 * i->YRR) - m) * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
}
//Draw a rectangular overlay
void draw_rectangular_overlay(SDL_Surface *s, struct _input *i){
	if(i->rotation_angle & 1){
		draw_rectangular_overlay_r(s, i);
		return;
	}
	int x,y,p,n,m;
	int pixel_index;
	int pixel_offset;
	int max_addr = (s->h + 1) * s->pitch;
	uint8_t *pixels = (uint8_t*)s->pixels;
	int xbase = (i->x - i->w/2) * i->XRR;
	int ybase = (i->y - i->h2/2) * i->YRR;
	if(xbase < 0)xbase = 0;
	if(ybase < 0)ybase = 0;
	y = ybase;
	m = min(5, i->h2*i->YRR);
	
	for(x = xbase;x < i->w*i->XRR+xbase && y < s->h;x++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = ((int)((i->h2 * i->YRR) + p)) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	x = xbase;
	m = min(5, i->w*i->XRR);
	for(y = ybase;y < i->h2*i->YRR + ybase;y++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = (p + (i->w * i->XRR) - m) * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
}
//Draw a trapezoidal overlay (one for each rotation)
void draw_trapezoidal_overlay_180(SDL_Surface *s, struct _input *i){
int x,y,p,n,m;
	int pixel_index;
	int pixel_offset;
	int max_addr = (s->h + 1) * s->pitch;
	uint8_t *pixels = (uint8_t*)s->pixels;
	int xbase = (i->x - i->w/2)*i->XRR;
	int ybase = (i->y - i->h1/2)*i->YRR;
	y = ybase;
	m = min(5, i->h1*i->YRR);
	for(x = xbase;x < (i->w*i->XRR)+xbase && y < s->h;x++){
		pixel_index =  x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			int h = (i->h2 + (double)(i->h1 - i->h2)*(double)((double)(x-xbase)/(double)(i->w*i->XRR)));
			int sy = (i->h1-h)/2 * i->YRR + y;
		//	printf("H = %d SY = %d xp = %0.2f\n", h, sy,(double)((double)(x-(i->x*i->XRR))/(double)(i->w*i->XRR)));
			pixel_offset = (p + sy) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = ((int)((h * i->YRR) + p + sy)) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	m = min(5, i->w*i->XRR);
	x = xbase;
	for(y = ybase;y < i->h1*i->YRR+ybase;y++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = (p + (i->w * i->XRR) - m) * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	int sy = (i->h1-i->h2)/2 * i->YRR;
	for(y = ybase;y < i->h2*i->YRR+ybase;y++){
		pixel_index = ((sy+y) * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p*3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
}
void draw_trapezoidal_overlay_90(SDL_Surface *s, struct _input *i){
	int x,y,p,n,m;
	int sx,w;
	int pixel_index;
	int pixel_offset;
	int max_addr = (s->h + 1) * s->pitch;
	uint8_t *pixels = (uint8_t*)s->pixels;
	int ybase = i->y*i->YRR - i->w*i->XRR/2;
	int xbase = i->x*i->XRR - i->h2*i->YRR/2;
	y = ybase;
	m = min(5, i->h2*i->YRR);
	for(x = xbase;x < (i->h2*i->YRR)+xbase && y < s->h;x++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		sx = (i->x - i->h1/2)*i->XRR;
		for(p = 0;p < m;p++){
			pixel_offset = p * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			if(x > sx && x < sx+(i->h1*i->XRR)){
				overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			}
			pixel_offset = ((int)((i->w * i->XRR) + p)) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	m = min(5, i->w*i->XRR);
	for(y = ybase;y < (i->w*i->XRR)+ybase;y++){
		w = (i->h1 + ((double)(i->h2 - i->h1) * ((double)(y - ybase)/(double)(i->w*i->XRR)))) * i->YRR;
		sx = i->x*i->XRR - w/2;
		debug("w = %d sx=%d\n", w, sx);
		x = sx; 
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = (p + w - m) * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
}
void draw_trapezoidal_overlay_270(SDL_Surface *s, struct _input *i){
int x,y,p,n,m;
	int sx,w;
	int pixel_index;
	int pixel_offset;
	int max_addr = (s->h + 1) * s->pitch;
	uint8_t *pixels = (uint8_t*)s->pixels;
	int h2 = i->h1;
	int h1 = i->h2;
	int ybase = i->y*i->YRR - i->w*i->XRR/2;
	int xbase = i->x*i->XRR - h1*i->YRR/2;
	y = ybase;
	m = min(5, h2*i->YRR);
	for(x = xbase;x < (h1*i->YRR)+xbase && y < s->h;x++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		sx = (i->x - h2/2)*i->XRR;
		for(p = 0;p < m;p++){
			//pixel_offset = ((int)((i->w * i->XRR) + p)) * s->pitch;
			pixel_offset = p * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			if(x > sx && x < sx+(h2*i->XRR)){
				pixel_offset = ((int)((i->w * i->XRR) + p)) * s->pitch;
				if(pixel_index + pixel_offset > max_addr) continue;
				overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			}
		}
	}
	m = min(5, i->w*i->XRR);
	for(y = ybase;y < (i->w*i->XRR)+ybase;y++){
		w = (h1 + ((double)(h2 - h1) * ((double)(y - ybase)/(double)(i->w*i->XRR)))) * i->YRR;
		sx = i->x*i->XRR - w/2;
		debug("w = %d sx=%d\n", w, sx);
		x = sx; 
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = (p + w - m) * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
}
void draw_trapezoidal_overlay(SDL_Surface *s, struct _input *i){
	if(i->rotation_angle == 1){ //90
		draw_trapezoidal_overlay_90(s, i);
		return;
	}
	if(i->rotation_angle == 2){ //180
		draw_trapezoidal_overlay_180(s, i);
		return;
	}
	if(i->rotation_angle == 3){ //270
		draw_trapezoidal_overlay_270(s, i);
		return;
	}
	//0
	int x,y,p,n,m;
	int pixel_index;
	int pixel_offset;
	int max_addr = (s->h + 1) * s->pitch;
	uint8_t *pixels = (uint8_t*)s->pixels;
	int xbase = (i->x - i->w/2)*i->XRR;
	int ybase = (i->y - i->h2/2)*i->YRR;
	y = ybase;
	m = min(5, i->h2*i->YRR);
	for(x = xbase;x < (i->w*i->XRR)+xbase && y < s->h;x++){
		pixel_index =  x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			int h = (i->h1 + (double)(i->h2 - i->h1)*(double)((double)(x-xbase)/(double)(i->w*i->XRR)));
			int sy = (i->h2-h)/2 * i->YRR + y;
		//	printf("H = %d SY = %d xp = %0.2f\n", h, sy,(double)((double)(x-(i->x*i->XRR))/(double)(i->w*i->XRR)));
			pixel_offset = (p + sy) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
			pixel_offset = ((int)((h * i->YRR) + p + sy)) * s->pitch;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	m = min(5, i->w*i->XRR);
	x = xbase;
	for(y = ybase;y < i->h2*i->YRR+ybase;y++){
		pixel_index = (y * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = (p + (i->w * i->XRR) - m) * 3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
	int sy = (i->h2-i->h1)/2 * i->YRR;
	for(y = ybase;y < i->h1*i->YRR+ybase;y++){
		pixel_index = ((sy+y) * s->pitch) + x*s->format->BytesPerPixel;
		for(p = 0;p < m;p++){
			pixel_offset = p*3;
			if(pixel_index + pixel_offset > max_addr) continue;
			overlay_alter_pixels(pixels, pixel_index + pixel_offset);
		}
	}
}
void draw_annular_overlay(SDL_Surface *s, struct _input *i){}
void draw_overlay(SDL_Surface *s, struct _input *i){
	if(i->window_type == 0){
		draw_rectangular_overlay(s, i);
	}
	if(i->window_type == 1){
		draw_trapezoidal_overlay(s, i);
	}
	if(i->window_type == 2){
		draw_annular_overlay(s, i);
	}
}
/*Update the preview image*/
void update_preview(SDL_Surface *image, SDL_Surface *windowed, struct _input *i){
	SDL_Rect d;
	d.x = 0;
	d.y = 0;
/*	if(input_preview && format_changed){
		SDL_FreeSurface(input_preview);
		input_preview = NULL;
	}
	if(!input_preview){
		input_preview = SDL_DisplayFormat(image);
	} else {
		SDL_BlitSurface(image, NULL, input_preview, NULL);
	}*/
	SDL_FillRect(screen, NULL, 0);
	if(i->prescaled){
		SDL_BlitSurface(image, NULL, screen, NULL);
	} else {
		d.w = image->w;
		d.h = image->h;
		while(d.w*2 < (screen->w/2)+48 && d.h*2 < screen->h){
			d.w*=2;
			d.h*=2;
		}
		while(d.w > (screen->w/2)+48 || d.h > screen->h){
			d.w/=1.25;
			d.h/=1.25;
		}
		if(d.w & 3){d.w += (4 - (d.w&3));} //Align to the next highest multiple for 4
		if(d.h & 3){d.h += (4 - (d.h&3));}
		i->XRR = (double)d.w/(double)i->Xres;
		i->YRR = (double)d.h/(double)i->Yres;
		scale_blit(screen, &d, image, 0, 0, image->w, image->h);
	}
	draw_overlay(screen, i);
	d.x = screen->w/2 + 50;
	d.y = 0;
	d.h = screen->h/2;
	d.w = d.h*nbtv_fmt->aspect_ratio;
	scale_blit(screen, &d, windowed, 0, 0, windowed->w, windowed->h);
	SDL_Flip(screen);
}
void update_input_preview(SDL_Surface *image, struct _input *i){
	SDL_BlitSurface(image, NULL, screen, NULL);
	draw_overlay(screen, i);
	SDL_Flip(screen);
}
/*User mouse click to change position*/
void handle_mouse_repos(struct _input *ip, int x, int y){
	int xp,yp;
	xp = x / ip->XRR;
	yp = y / ip->YRR;
	if(xp > ip->Xres || yp > ip->Yres){
		return;
	}
	if(ip->window_type == 2){
		return;
	}
	if(xp < ip->Xres){
		if(ip->rotation_angle & 1){
			if(xp + ip->h2/2 > ip->Xres){
				xp = ip->Xres - ip->h2/2;
			}
			if(xp < ip->h2/2){
				xp = ip->h2/2;
			}
		} else {
			if(xp + ip->w/2 > ip->Xres){
				xp = ip->Xres - ip->w/2;
			}
			if(xp < ip->w/2){
				xp = ip->w/2;
			}
		}
		ip->x = xp;
	}
	if(yp < ip->Yres){
		if(ip->rotation_angle & 1){
			if(yp + ip->w/2 > ip->Yres){
				yp = ip->Yres - ip->w/2;
			}
			if(yp < ip->w/2){
				yp = ip->w/2;
			}
		} else {
			if(yp + ip->h2/2 > ip->Yres){
				yp = ip->Yres - ip->h2/2;
			}
			if(yp < ip->h2/2){
				yp = ip->h2/2;
			}
		}
		ip->y = yp;
	}
	gui_update_input();
	format_changed = 1;
}